module Schleuder
  require 'logger'
  # Extends Logger, adds notification sending
  class SchleuderLogger < Logger
    # instantiates a Logger and sets @+progname+ and @+sev_threshold+
    def initialize(progname, logfile, loglevel="warn")
      logfile = nil if logfile == 'syslog'
      if Schleuder.config.log_rotate_keep > 0 && ! logfile.nil? then
        # rotate daily
        super logfile, 'daily'
      else
        super logfile
      end
      @progname = progname
      if logfile.nil?
        init_syslog
      end
      # By specifying the log-level we can define what messages are actually
      # logged.  For 'ageing' logfiles and/or set a maximum size see
      # documentation for +Logger+.
      @sev_threshold = eval("Logger::" + loglevel.upcase)

      if Schleuder.config.log_rotate_keep > 0 && ! logfile.nil? then
        # delete out-of-age logfiles
        removed = File.unlink(*Dir["#{file}.20[0-9]*"].sort.slice(0..-Schleuder.config.log_rotate_keep))
        @log.info "Removed #{removed} old logfiles" if removed > 0
      end
    end

    def init_syslog
        require 'syslog'
        require 'stringio'

        unless Syslog.opened?
	  @syslog = Syslog.open(@progname, Syslog::LOG_PID | Syslog::LOG_NDELAY,
                                Syslog::LOG_MAIL)
        else
	  @syslog = Syslog.reopen(@progname, Syslog::LOG_PID | Syslog::LOG_NDELAY,
                                  Syslog::LOG_MAIL)
        end
        @@LOGGER_LEVEL_TO_SYSLOG = {
          Logger::DEBUG   => Syslog::LOG_DEBUG,
          Logger::INFO    => Syslog::LOG_INFO,
          Logger::WARN    => Syslog::LOG_WARNING,
          Logger::ERROR   => Syslog::LOG_ERR,
          Logger::FATAL   => Syslog::LOG_CRIT,
          Logger::UNKNOWN => Syslog::LOG_NOTICE,
        }
	# On top of using syslog(3), continue to use Logger to keep a backlog
        # handy for notify_admin
        @logdev = Logger::LogDevice.new(StringIO.new)
    end

    def add(severity, message = nil, progname = nil, &block)
      severity ||= Logger::UNKNOWN
      return true if severity < @level
      if message.nil?
        if block_given?
          message = yield
        else
          message = progname
        end
      end

      super(severity, message)
      return true unless @syslog
      @syslog.log(@@LOGGER_LEVEL_TO_SYSLOG[severity], '%s', message)
    end

    # extends Logger#error by sending notification to admin
    def error(*msg)
      super(*msg)
      notify_admin unless Schleuder.list.nil?
    end

    # extends Logger#fatal by sending notification to admin and superadmin,
    # closing self and exiting (actually dieing is what "fatal" means, no?)
    def fatal(*msg)
      super(*msg)
      notify_admin unless Schleuder.list.nil?
      self.close
      exit 1
    end

    def read_backlog
      unless @syslog
        logfile = File.expand_path(@logdev.instance_variable_get('@filename'))
        File.read(logfile)
      else
        @logdev.dev.string
      end
    end

    def notify_admin msg=nil, subject=nil
      self.info 'notifying (super)admin'
      m = Mail.new
      m.subject = subject || "Error"
      m.date = Time.now
      m.from = Schleuder.list.config.myaddr

      if msg.nil?
        msg = "Hello,\n\nan error has occurred working for list #{Schleuder.list.listname}:\n\n"
        read_backlog.split(/\n[IDEFW]{1}, /).each do |line|
          msg << "#{line}\n" if line.match(/ ##{$$}\]/)
        end
      end

      m.body = msg
      altmsg = "Hello,\n\nsending an encrypted error message to you failed. Therefore you receive only\nthis message and are kindly requested to take care of the encryption problem\n(e.g. fix your keys) and have a look at the logs to find the error.\n\nYours,  Schleuder\n"

      begin
        if Schleuder.list.config.adminaddr.kind_of?(Array) and not Schleuder.list.config.adminaddr.empty?
          Schleuder.list.config.adminaddr.each do |addr|
            r = Member.new('email' => addr)
            mail = m.individualize(r)
            if not Processor.send(mail, r)
              mail.body = altmsg
              Processor.send(mail, r, false)
            end
          end
        else
          m.to = Schleuder.config.superadminaddr
          m.body = "[Working for list #{Schleuder.list.listname} I found adminaddr is not (correctly) set.\nPlease fix this!\n\nThe errormessage is appended.]\n\n#{m.body}"
          Processor.send(m, Member.new('email' => Schleuder.config.superadminaddr))
        end
      rescue => e
        puts "Something's seriously wrong: #{e}\n#{e.backtrace.join("\n")}\n\nThe message:\n#{m.to_s}"
        exit 1
      end
    end
  end
end
